@@ -47,6 +47,7 @@ struct connman_service {
 	connman_bool_t favorite;
 	connman_bool_t hidden;
 	connman_bool_t ignore;
+	connman_bool_t autoconnect;
 	GTimeVal modified;
 	unsigned int order;
 	char *name;
@@ -290,12 +291,91 @@ static void strength_changed(struct connman_service *service)
 	g_dbus_send_message(connection, signal);
 }
 
+static void passphrase_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	dbus_bool_t required;
+	const char *key = "PassphraseRequired";
+
+	if (service->path == NULL)
+		return;
+
+	switch (service->type) {
+	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+	case CONNMAN_SERVICE_TYPE_ETHERNET:
+	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		return;
+	case CONNMAN_SERVICE_TYPE_WIFI:
+		required = FALSE;
+
+		switch (service->security) {
+		case CONNMAN_SERVICE_SECURITY_UNKNOWN:
+		case CONNMAN_SERVICE_SECURITY_NONE:
+			break;
+		case CONNMAN_SERVICE_SECURITY_WEP:
+		case CONNMAN_SERVICE_SECURITY_WPA:
+		case CONNMAN_SERVICE_SECURITY_RSN:
+			if (service->passphrase == NULL)
+				required = TRUE;
+			break;
+		}
+		break;
+	}
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN, &required);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
+static void autoconnect_changed(struct connman_service *service)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *key = "AutoConnect";
+
+	if (service->path == NULL)
+		return;
+
+	signal = dbus_message_new_signal(service->path,
+				CONNMAN_SERVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN,
+							&service->autoconnect);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+}
+
 static DBusMessage *get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
 	struct connman_service *service = user_data;
 	DBusMessage *reply;
 	DBusMessageIter array, dict;
+	dbus_bool_t required;
 	const char *str;
 
 	DBG("service %p", service);
@@ -343,17 +423,53 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	connman_dbus_dict_append_variant(&dict, "Favorite",
 					DBUS_TYPE_BOOLEAN, &service->favorite);
 
+	if (service->favorite == TRUE)
+		connman_dbus_dict_append_variant(&dict, "AutoConnect",
+				DBUS_TYPE_BOOLEAN, &service->autoconnect);
+	else
+		connman_dbus_dict_append_variant(&dict, "AutoConnect",
+					DBUS_TYPE_BOOLEAN, &service->favorite);
+
 	if (service->name != NULL)
 		connman_dbus_dict_append_variant(&dict, "Name",
 					DBUS_TYPE_STRING, &service->name);
 
-	if (service->passphrase != NULL &&
-			__connman_security_check_privilege(msg,
-				CONNMAN_SECURITY_PRIVILEGE_SECRET) == 0)
-		connman_dbus_dict_append_variant(&dict, "Passphrase",
+	switch (service->type) {
+	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+	case CONNMAN_SERVICE_TYPE_ETHERNET:
+	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		break;
+	case CONNMAN_SERVICE_TYPE_WIFI:
+		if (service->passphrase != NULL &&
+				__connman_security_check_privilege(msg,
+					CONNMAN_SECURITY_PRIVILEGE_SECRET) == 0)
+			connman_dbus_dict_append_variant(&dict, "Passphrase",
 				DBUS_TYPE_STRING, &service->passphrase);
 
-	__connman_ipconfig_append_ipv4(service->ipconfig, &dict, "IPv4.");
+		required = FALSE;
+
+		switch (service->security) {
+		case CONNMAN_SERVICE_SECURITY_UNKNOWN:
+		case CONNMAN_SERVICE_SECURITY_NONE:
+			break;
+		case CONNMAN_SERVICE_SECURITY_WEP:
+		case CONNMAN_SERVICE_SECURITY_WPA:
+		case CONNMAN_SERVICE_SECURITY_RSN:
+			if (service->passphrase == NULL)
+				required = TRUE;
+			break;
+		}
+
+		connman_dbus_dict_append_variant(&dict, "PassphraseRequired",
+						DBUS_TYPE_BOOLEAN, &required);
+		break;
+	}
+
+	if (service->ipconfig != NULL)
+		__connman_ipconfig_append_ipv4(service->ipconfig,
+							&dict, "IPv4.");
 
 	dbus_message_iter_close_container(&array, &dict);
 
@@ -398,14 +514,38 @@ static DBusMessage *set_property(DBusConnection *conn,
 		g_free(service->passphrase);
 		service->passphrase = g_strdup(passphrase);
 
+		passphrase_changed(service);
+
 		if (service->network != NULL)
 			connman_network_set_string(service->network,
 				"WiFi.Passphrase", service->passphrase);
 
+		__connman_storage_save_service(service);
+	} else if (g_str_has_prefix(name, "AutoConnect") == TRUE) {
+		connman_bool_t autoconnect;
+
+		if (type != DBUS_TYPE_BOOLEAN)
+			return __connman_error_invalid_arguments(msg);
+
+		if (service->favorite == FALSE)
+			return __connman_error_invalid_service(msg);
+
+		dbus_message_iter_get_basic(&value, &autoconnect);
+
+		if (service->autoconnect == autoconnect)
+			return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
+
+		service->autoconnect = autoconnect;
+
+		autoconnect_changed(service);
+
 		__connman_storage_save_service(service);
 	} else if (g_str_has_prefix(name, "IPv4.") == TRUE) {
 		int err;
 
+		if (service->ipconfig == NULL)
+			return __connman_error_invalid_property(msg);
+
 		err = __connman_ipconfig_set_ipv4(service->ipconfig,
 							name + 5, &value);
 		if (err < 0)
@@ -442,6 +582,13 @@ static DBusMessage *clear_property(DBusConnection *conn,
 		set_idle(service);
 
 		g_get_current_time(&service->modified);
+		__connman_storage_save_service(service);
+	} else if (g_str_equal(name, "Passphrase") == TRUE) {
+		g_free(service->passphrase);
+		service->passphrase = NULL;
+
+		passphrase_changed(service);
+
 		__connman_storage_save_service(service);
 	} else
 		return __connman_error_invalid_property(msg);
@@ -469,6 +616,9 @@ static connman_bool_t is_connecting(struct connman_service *service)
 
 static connman_bool_t is_ignore(struct connman_service *service)
 {
+	if (service->autoconnect == FALSE)
+		return TRUE;
+
 	if (service->ignore == TRUE)
 		return TRUE;
 
@@ -548,29 +698,12 @@ static gboolean connect_timeout(gpointer user_data)
 
 	service->timeout = 0;
 
-	if (service->network != NULL) {
-		connman_bool_t connected;
-
-		connected = connman_network_get_connected(service->network);
-		if (connected == TRUE) {
-			__connman_service_indicate_state(service,
-						CONNMAN_SERVICE_STATE_READY);
-			return FALSE;
-		}
-
+	if (service->network != NULL)
 		__connman_network_disconnect(service->network);
-	} else if (service->device != NULL) {
-		connman_bool_t disconnected;
-
-		disconnected = connman_device_get_disconnected(service->device);
-		if (disconnected == FALSE) {
-			__connman_service_indicate_state(service,
-						CONNMAN_SERVICE_STATE_READY);
-			return FALSE;
-		}
-
+	else if (service->device != NULL)
 		__connman_device_disconnect(service->device);
-	}
+
+	__connman_ipconfig_disable(service->ipconfig);
 
 	if (service->pending != NULL) {
 		DBusMessage *reply;
@@ -679,6 +812,8 @@ static DBusMessage *remove_service(DBusConnection *conn,
 	g_free(service->passphrase);
 	service->passphrase = NULL;
 
+	passphrase_changed(service);
+
 	connman_service_set_favorite(service, FALSE);
 	__connman_storage_save_service(service);
 
@@ -707,6 +842,9 @@ static DBusMessage *move_before(DBusConnection *conn,
 
 	DBG("target %s", target->identifier);
 
+	if (target->state != service->state)
+		return __connman_error_invalid_service(msg);
+
 	g_get_current_time(&service->modified);
 	__connman_storage_save_service(service);
 
@@ -741,6 +879,9 @@ static DBusMessage *move_after(DBusConnection *conn,
 
 	DBG("target %s", target->identifier);
 
+	if (target->state != service->state)
+		return __connman_error_invalid_service(msg);
+
 	g_get_current_time(&service->modified);
 	__connman_storage_save_service(service);
 
@@ -789,7 +930,10 @@ static void service_free(gpointer user_data)
 	if (service->network != NULL)
 		connman_network_unref(service->network);
 
-	connman_ipconfig_unref(service->ipconfig);
+	if (service->ipconfig != NULL) {
+		connman_ipconfig_unref(service->ipconfig);
+		service->ipconfig = NULL;
+	}
 
 	g_free(service->profile);
 	g_free(service->name);
@@ -858,15 +1002,6 @@ struct connman_service *connman_service_create(void)
 
 	__connman_service_initialize(service);
 
-	service->ipconfig = connman_ipconfig_create();
-	if (service->ipconfig == NULL) {
-		g_free(service);
-		return NULL;
-	}
-
-	connman_ipconfig_set_method(service->ipconfig,
-					CONNMAN_IPCONFIG_METHOD_DHCP);
-
 	return service;
 }
 
@@ -919,6 +1054,20 @@ static gint service_compare(gconstpointer a, gconstpointer b,
 	if (service_a->favorite == FALSE && service_b->favorite == TRUE)
 		return 1;
 
+	if (service_a->type != service_b->type) {
+		switch (service_a->type) {
+		case CONNMAN_SERVICE_TYPE_UNKNOWN:
+		case CONNMAN_SERVICE_TYPE_ETHERNET:
+			break;
+		case CONNMAN_SERVICE_TYPE_WIFI:
+			return 1;
+		case CONNMAN_SERVICE_TYPE_WIMAX:
+		case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+		case CONNMAN_SERVICE_TYPE_CELLULAR:
+			return -1;
+		}
+	}
+
 	return (gint) service_b->strength - (gint) service_a->strength;
 }
 
@@ -1109,22 +1258,51 @@ int __connman_service_connect(struct connman_service *service)
 	if (is_connecting(service) == TRUE)
 		return -EALREADY;
 
+	switch (service->type) {
+	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+		return -EINVAL;
+	case CONNMAN_SERVICE_TYPE_ETHERNET:
+	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		break;
+	case CONNMAN_SERVICE_TYPE_WIFI:
+		switch (service->security) {
+		case CONNMAN_SERVICE_SECURITY_UNKNOWN:
+		case CONNMAN_SERVICE_SECURITY_NONE:
+			break;
+		case CONNMAN_SERVICE_SECURITY_WEP:
+		case CONNMAN_SERVICE_SECURITY_WPA:
+		case CONNMAN_SERVICE_SECURITY_RSN:
+			if (service->passphrase == NULL)
+				return -ENOKEY;
+			break;
+		}
+		break;
+	}
+
 	if (service->network != NULL) {
 		if (prepare_network(service) == FALSE)
 			return -EINVAL;
 
+		__connman_ipconfig_enable(service->ipconfig);
+
 		err = __connman_network_connect(service->network);
 	} else if (service->device != NULL) {
 		if (service->favorite == FALSE)
 			return -ENOLINK;
 
+		__connman_ipconfig_enable(service->ipconfig);
+
 		err = __connman_device_connect(service->device);
 	} else
 		return -EOPNOTSUPP;
 
 	if (err < 0) {
-		if (err != -EINPROGRESS)
+		if (err != -EINPROGRESS) {
+			__connman_ipconfig_disable(service->ipconfig);
 			return err;
+		}
 
 		service->timeout = g_timeout_add_seconds(45,
 						connect_timeout, service);
@@ -1152,6 +1330,8 @@ int __connman_service_disconnect(struct connman_service *service)
 	} else
 		return -EOPNOTSUPP;
 
+	__connman_ipconfig_disable(service->ipconfig);
+
 	if (err < 0) {
 		if (err != -EINPROGRESS)
 			return err;
@@ -1294,6 +1474,12 @@ int __connman_service_create_and_connect(DBusMessage *msg)
 	if (ssid_len < 1)
 		return -EINVAL;
 
+	if (g_strcmp0(security, "none") != 0 &&
+			g_strcmp0(security, "wep") != 0 &&
+				g_strcmp0(security, "wpa") != 0 &&
+					g_strcmp0(security, "rsn") != 0)
+		return -EINVAL;
+
 	device = __connman_element_find_device(CONNMAN_SERVICE_TYPE_WIFI);
 	if (device == NULL)
 		return -EOPNOTSUPP;
@@ -1432,6 +1618,65 @@ static int service_register(struct connman_service *service)
 	return 0;
 }
 
+static void service_up(struct connman_ipconfig *ipconfig)
+{
+	connman_info("%s up", connman_ipconfig_get_ifname(ipconfig));
+}
+
+static void service_down(struct connman_ipconfig *ipconfig)
+{
+	connman_info("%s down", connman_ipconfig_get_ifname(ipconfig));
+}
+
+static void service_lower_up(struct connman_ipconfig *ipconfig)
+{
+	connman_info("%s lower up", connman_ipconfig_get_ifname(ipconfig));
+}
+
+static void service_lower_down(struct connman_ipconfig *ipconfig)
+{
+	connman_info("%s lower down", connman_ipconfig_get_ifname(ipconfig));
+}
+
+static void service_ip_bound(struct connman_ipconfig *ipconfig)
+{
+	connman_info("%s ip bound", connman_ipconfig_get_ifname(ipconfig));
+}
+
+static void service_ip_release(struct connman_ipconfig *ipconfig)
+{
+	connman_info("%s ip release", connman_ipconfig_get_ifname(ipconfig));
+}
+
+static const struct connman_ipconfig_ops service_ops = {
+	.up		= service_up,
+	.down		= service_down,
+	.lower_up	= service_lower_up,
+	.lower_down	= service_lower_down,
+	.ip_bound	= service_ip_bound,
+	.ip_release	= service_ip_release,
+};
+
+static void setup_ipconfig(struct connman_service *service, int index)
+{
+	if (index < 0)
+		return;
+
+	service->ipconfig = connman_ipconfig_create(index);
+	if (service->ipconfig == NULL)
+		return;
+
+	connman_ipconfig_set_method(service->ipconfig,
+					CONNMAN_IPCONFIG_METHOD_DHCP);
+
+	__connman_storage_load_service(service);
+
+	connman_ipconfig_set_data(service->ipconfig, service);
+
+	connman_ipconfig_set_ops(service->ipconfig, &service_ops);
+	connman_ipconfig_set_ops(service->ipconfig, NULL);
+}
+
 /**
  * __connman_service_lookup_from_device:
  * @device: device structure
@@ -1487,8 +1732,12 @@ struct connman_service *__connman_service_create_from_device(struct connman_devi
 
 	service->type = __connman_device_get_service_type(device);
 
+	service->autoconnect = FALSE;
+
 	service->device = device;
 
+	setup_ipconfig(service, connman_device_get_index(device));
+
 	service_register(service);
 
 	__connman_profile_changed(TRUE);
@@ -1578,16 +1827,17 @@ static enum connman_service_type convert_network_type(struct connman_network *ne
 	switch (type) {
 	case CONNMAN_NETWORK_TYPE_UNKNOWN:
 	case CONNMAN_NETWORK_TYPE_VENDOR:
-	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
-	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
 		break;
-	case CONNMAN_NETWORK_TYPE_MBM:
-	case CONNMAN_NETWORK_TYPE_HSO:
-		return CONNMAN_SERVICE_TYPE_CELLULAR;
 	case CONNMAN_NETWORK_TYPE_WIFI:
 		return CONNMAN_SERVICE_TYPE_WIFI;
 	case CONNMAN_NETWORK_TYPE_WIMAX:
 		return CONNMAN_SERVICE_TYPE_WIMAX;
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
+		return CONNMAN_SERVICE_TYPE_BLUETOOTH;
+	case CONNMAN_NETWORK_TYPE_MBM:
+	case CONNMAN_NETWORK_TYPE_HSO:
+		return CONNMAN_SERVICE_TYPE_CELLULAR;
 	}
 
 	return CONNMAN_SERVICE_TYPE_UNKNOWN;
@@ -1707,10 +1957,25 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 
 	service->type = convert_network_type(network);
 
+	switch (service->type) {
+	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+	case CONNMAN_SERVICE_TYPE_ETHERNET:
+	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
+		service->autoconnect = FALSE;
+		break;
+	case CONNMAN_SERVICE_TYPE_WIFI:
+		service->autoconnect = TRUE;
+		break;
+	}
+
 	service->state = CONNMAN_SERVICE_STATE_IDLE;
 
 	update_from_network(service, network);
 
+	setup_ipconfig(service, connman_network_get_index(network));
+
 	service_register(service);
 
 	__connman_profile_changed(TRUE);
@@ -1721,6 +1986,27 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 	return service;
 }
 
+void __connman_service_update_from_network(struct connman_network *network)
+{
+	struct connman_service *service;
+	connman_uint8_t strength;
+
+	service = __connman_service_lookup_from_network(network);
+	if (service == NULL)
+		return;
+
+	if (service->network == NULL)
+		return;
+
+	strength = connman_network_get_uint8(service->network, "Strength");
+	if (strength == service->strength)
+		return;
+
+	service->strength = strength;
+
+	strength_changed(service);
+}
+
 void __connman_service_remove_from_network(struct connman_network *network)
 {
 	struct connman_service *service;
@@ -1734,6 +2020,7 @@ void __connman_service_remove_from_network(struct connman_network *network)
 
 static int service_load(struct connman_service *service)
 {
+	const char *ident = service->profile;
 	GKeyFile *keyfile;
 	gchar *pathname, *data = NULL;
 	gsize length;
@@ -1743,10 +2030,10 @@ static int service_load(struct connman_service *service)
 
 	DBG("service %p", service);
 
-	if (service->profile == NULL)
+	if (ident == NULL)
 		return -EINVAL;
 
-	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR, service->profile);
+	pathname = g_strdup_printf("%s/%s.profile", STORAGEDIR, ident);
 	if (pathname == NULL)
 		return -ENOMEM;
 
@@ -1827,6 +2114,9 @@ static int service_load(struct connman_service *service)
 		service->favorite = g_key_file_get_boolean(keyfile,
 				service->identifier, "Favorite", NULL);
 
+		service->autoconnect = g_key_file_get_boolean(keyfile,
+				service->identifier, "AutoConnect", NULL);
+
 		str = g_key_file_get_string(keyfile,
 				service->identifier, "Failure", NULL);
 		if (str != NULL) {
@@ -1850,7 +2140,8 @@ static int service_load(struct connman_service *service)
 		service->passphrase = str;
 	}
 
-	__connman_ipconfig_load(service->ipconfig, keyfile,
+	if (service->ipconfig != NULL)
+		__connman_ipconfig_load(service->ipconfig, keyfile,
 					service->identifier, "IPv4.");
 
 done:
@@ -1861,6 +2152,7 @@ done:
 
 static int service_save(struct connman_service *service)
 {
+	const char *ident = service->profile;
 	GKeyFile *keyfile;
 	gchar *pathname, *data = NULL;
 	gsize length;
@@ -1869,10 +2161,10 @@ static int service_save(struct connman_service *service)
 
 	DBG("service %p", service);
 
-	if (service->profile == NULL)
+	if (ident == NULL)
 		return -EINVAL;
 
-	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR, service->profile);
+	pathname = g_strdup_printf("%s/%s.profile", STORAGEDIR, ident);
 	if (pathname == NULL)
 		return -ENOMEM;
 
@@ -1935,6 +2227,10 @@ update:
 		g_key_file_set_boolean(keyfile, service->identifier,
 					"Favorite", service->favorite);
 
+		if (service->favorite == TRUE)
+			g_key_file_set_boolean(keyfile, service->identifier,
+					"AutoConnect", service->autoconnect);
+
 		if (service->state == CONNMAN_SERVICE_STATE_FAILURE) {
 			const char *failure = error2string(service->error);
 			if (failure != NULL)
@@ -1962,7 +2258,8 @@ update:
 		g_key_file_remove_key(keyfile, service->identifier,
 							"Passphrase", NULL);
 
-	__connman_ipconfig_save(service->ipconfig, keyfile,
+	if (service->ipconfig != NULL)
+		__connman_ipconfig_save(service->ipconfig, keyfile,
 					service->identifier, "IPv4.");
 
 	data = g_key_file_to_data(keyfile, &length, NULL);
